from pathlib import Path
from typing import TypedDict
import csv


class Option(TypedDict):
    chord: str
    score: int


class Chord(TypedDict):
    word: str
    chord: str
    reserved_chord: str
    type: str
    alt1: str
    alt2: str
    alt3: str
    options: list[Option] | None


def load_file(file: Path) -> list[Chord]:
    with open(file) as f:
        reader = csv.DictReader(f)
        chords: list[Chord] = [line for line in reader]
        return chords


def validate_chords(chords: list[Chord]):
    used = {}

    for chord in chords:
        c = chord["chord"]
        if c:
            sorted_chord = "".join(sorted(c))
            if sorted_chord in used:
                raise Exception(
                    f"Error: chord '{c}' for word {chord['word']} already used by {used[sorted_chord]}"
                )

            used[sorted_chord] = chord["word"]



def find_combinations(s, prefix="", index=0):
    """
    Recursively find every combination of letters in a string from left to right
    starting with the first character.

    Args:
    - s: The input string.
    - prefix: The current combination being built.
    - index: The current index in the string.

    Returns:
    - result: A list containing all combinations starting with the first character
              and excluding the empty string, sorted by shortest string first.
    """
    result = []
    if index == len(s):
        if prefix:
            result.append(prefix)
        return result

    # Include the current character only if it's the first character or part of a previous combination
    if index == 0 or prefix:
        result.extend(find_combinations(s, prefix + s[index], index + 1))

    # Exclude the current character
    result.extend(find_combinations(s, prefix, index + 1))

    return result


def find_all_combinations(string):
    if len(string) == 1:
        return [string]
    else:
        chords = []
        for i, char in enumerate(string):
            remaining_chars = string[:i] + string[i + 1 :]
            sub_chords = find_all_combinations(remaining_chars)
            for sub_chord in sub_chords:
                chords.append(char + sub_chord)
        return chords



from chordgen.chord import Chord
from chordgen.constants import CONFIG_DIR
from pydantic import BaseModel, Field
from chordgen.pydantic import File

key_map = {
    "'": "QUOT",
    ";": "SEMI",
    ",": "COMMA",
    ".": "DOT",
    "-": "MINUS",
    " ": "SPC",
    "@": "AT",
    "?": "QUESTION",
    "←": "BSPC",
}


class ZmkOutput(BaseModel):
    chords_file: File = CONFIG_DIR / "zmk_chords.dtsi"
    macros_file: File = CONFIG_DIR / "zmk_macros.dtsi"
    chord_keys: list[str] = ["$"]
    shifted_chord_keys: list[str] = ["$", "#"]
    alt1_keys: list[str] = ["!"]
    alt2_keys: list[str] = ["@"]
    alt3_keys: list[str] = ["!", "@"]
    limit: int = Field(
        200,
        description="Limit the number of chords outputted. It will vary depending on MCU, but I have found ZMK to be quite limited compared to QMK with the number or chords it can fit.",
    )
    chord_timeout: int = 100

    key_positions: list[str] = Field(
        default=[
            "_qwertyuiop_",
            "_asdfghjkl;_",
            "_zxcvbnm,./_",
            "_!@#$_",
        ],
        description="Should contain the same number of characters as keys on your keyboard. Use _ to ignore a key. Chord, shift, alt1, alt2 can be represented by shifted characters such as $#!@. Spaces will be ignored and can be used for formatting.",
    )

    def output(self, chords: list[Chord]):
        key_positions = [
            item for items in self.key_positions for item in items if item != " "
        ]
        macros_output = """#define MACRO(NAME, BINDINGS) \\
        macro_##NAME: macro_##NAME { \\
            compatible = "zmk,behavior-macro"; \\
            #binding-cells = <0>; \\
            wait-ms = <0>; \\
            tap-ms = <10>; \\
            bindings = <BINDINGS>; \\
        };

"""

        chords_output = (
            """#define CHORD(NAME, BINDINGS, KEYPOS) \\
        chord_##NAME { \\
            timeout-ms = <"""
            + str(self.chord_timeout)
            + """>; \\
            bindings = <BINDINGS>; \\
            key-positions = <KEYPOS>; \\
            layers = <0>; \\
        };

"""
        )

        key_positions_map = {}
        for i, key in enumerate(key_positions):
            key_positions_map[key] = i

        def translate_keys(chord):
            result = []
            for k in chord:
                if k in key_positions_map:
                    result.append(key_positions_map[k])
                else:
                    raise Exception(
                        f'Unable to find key position for {k}, is it in "key_positions"?'
                    )
            result = [str(i) for i in result]
            return result

        def translate_macro(word, capitalize=False):
            result = []
            for i, k in enumerate(word):
                k = k.upper()
                kp = "&kp "
                if capitalize and i == 0:
                    kp += "LS("

                if k in key_map:
                    kp += key_map[k]
                else:
                    kp += k

                if capitalize and i == 0:
                    kp += ")"

                result.append(kp)
            return result

        # Convienience macros for punctuation
        for p in [";", ",", "."]:
            if p not in key_positions_map:
                continue
            name = f"c_{key_map[p]}"
            macro = translate_macro(f"←{p} ")
            positions = translate_keys([p] + self.chord_keys)
            macros_output += f"MACRO({name}, {' '.join(macro)})\n"
            chords_output += f"CHORD({name}, &macro_{name}, {' '.join(positions)})\n"

        alt_keys = [self.alt1_keys, self.alt2_keys, self.alt3_keys]
        count = 0
        for chord in chords:
            c = chord["chord"]
            if not c:
                continue

            count += 1
            if self.limit != 0 and count > self.limit:
                print(f"Stopping at chord {self.limit} due to limit setting")
                break

            words = [chord["word"], chord["alt1"], chord["alt2"], chord["alt3"]]
            for i, word in enumerate(words):
                if not word:
                    continue
                alt = []
                if i > 0:
                    alt = alt_keys[i - 1]
                name = f"c_{c}{'_' * i}".replace("'", "_")
                macro = translate_macro(word + " ")

                positions = translate_keys(list(c) + self.chord_keys + alt)
                macros_output += f"MACRO({name}, {' '.join(macro)})\n"
                chords_output += (
                    f"CHORD({name}, &macro_{name}, {' '.join(positions)})\n"
                )

                if self.shifted_chord_keys:
                    positions = translate_keys(list(c) + self.shifted_chord_keys + alt)
                    macro = translate_macro(word + " ", True)
                    macros_output += f"MACRO(s_{name}, {' '.join(macro)})\n"
                    chords_output += (
                        f"CHORD(s_{name}, &macro_s_{name}, {' '.join(positions)})\n"
                    )

        print(f"Writing {self.macros_file}")
        with open(self.macros_file, "w") as file:
            file.write(macros_output)

        print(f"Writing {self.chords_file}")
        with open(self.chords_file, "w") as file:
            file.write(chords_output)



from chordgen.chord import Chord
from chordgen.constants import CONFIG_DIR
from pydantic import BaseModel
from chordgen.pydantic import File


qmk_key_codes = {
    ";": "KC_SCLN",
    ",": "KC_COMMA",
    ".": "KC_DOT",
    "'": "KC_QUOT",
    "-": "KC_MINUS",
}


class QmkOutput(BaseModel):
    file: File = CONFIG_DIR / "qmk_chords.def"
    chord_keys: list[str] = ["KC_CHORD"]
    shifted_chord_keys: list[str] = ["KC_CHORD", "KC_CHORD_SFT"]
    alt1_keys: list[str] = ["KC_CHORD_ALT1"]
    alt2_keys: list[str] = ["KC_CHORD_ALT2"]
    alt3_keys: list[str] = ["KC_CHORD_ALT1", "KC_CHORD_ALT2"]
    key_codes: dict[str, str] = {
        "A": "KC_SFT_A",
        "S": "KC_ALT_S",
        "D": "KC_GUI_D",
        "F": "KC_CTL_F",
        "J": "KC_CTL_J",
        "K": "KC_GUI_K",
        "L": "KC_ALT_L",
        ";": "KC_SFT_SEMI",
    }

    def translate_keys(self, chord):
        result = []
        key_codes = qmk_key_codes | self.key_codes

        for k in chord:
            k = k.upper()
            if k in key_codes:
                result.append(key_codes[k])
            elif k.isalnum():
                result.append(f"KC_{k}")
            else:
                raise Exception(
                    f"Unknown QMK code to map '{k}', add it to the chordgen config"
                )

        return result

    def output(self, chords: list[Chord]):
        alt_keys = [self.alt1_keys, self.alt2_keys, self.alt3_keys]
        output = ""
        for chord in chords:
            if not chord["chord"]:
                continue
            words = [chord["word"], chord["alt1"], chord["alt2"], chord["alt3"]]
            for i, word in enumerate(words):
                if not word:
                    continue
                keys = self.translate_keys(chord["chord"])
                alt = []
                if i > 0:
                    alt = alt_keys[i - 1]
                name = f"c_{chord['chord']}{i}".replace("'", "_").replace("-", "_")

                output += f'SUBS({name}, "{word} ", {", ".join(keys + self.chord_keys + alt)})\n'
                if self.shifted_chord_keys:
                    output += f'SUBS({name}s, "{word.capitalize()} ", {", ".join(keys + self.shifted_chord_keys + alt)})\n'

        print(f"Writing {self.file}")
        with open(self.file, "w") as file:
            file.write(output)



from pathlib import Path
from typing import Annotated
from pydantic import BeforeValidator, PlainSerializer


def serialize_path(value: Path) -> str:
    try:
        return f"~/{value.relative_to(Path.home())}"
    except ValueError:
        return str(value)


def expand_user(v):
    # Ensure "~" gets expanded if user provides it
    return Path(v).expanduser()


File = Annotated[
    Path, BeforeValidator(expand_user), PlainSerializer(serialize_path, return_type=str)
]



from chordgen.chord import Chord
from chordgen.constants import CONFIG_DIR
from pydantic import BaseModel
from chordgen.pydantic import File


class TrainingOutput(BaseModel):
    file: File = CONFIG_DIR / "training.txt"

    def output(self, chords: list[Chord]):
        index = 0
        line_words = ""
        line_chords = ""

        print(f"Writing {self.file}")
        with open(self.file, "w") as training_file:
            for chord in chords:
                c = chord["chord"]
                if not c:
                    continue
                word = chord["word"]
                line_words += f"{word} "
                line_chords += c + (" " * (len(word) + 1 - len(c)))
                index += 1
                if index % 10 == 0:
                    training_file.write(line_words.rstrip())
                    training_file.write("\n")
                    training_file.write(line_chords.rstrip())
                    training_file.write("\n")
                    line_words = ""
                    line_chords = ""



import json
from chordgen.chord import Chord
from chordgen.constants import CONFIG_DIR
from pydantic import BaseModel
from chordgen.pydantic import File


class CharaChorderOutput(BaseModel):
    file: File = CONFIG_DIR / "charachorder_chords.json"

    def translate_keys(self, chord: str):
        result = [ord(c) for c in chord]
        return result

    def output(self, chords: list[Chord]):
        output_chords = []
        for chord in chords:
            if not chord["chord"]:
                continue
            chord_keys = self.translate_keys(chord["chord"])
            word_keys = self.translate_keys(chord["word"])
            output_chords.append([chord_keys, word_keys])

        output = {"charaVersion": 1, "type": "chords", "chords": output_chords}
        print(f"Writing {self.file}")
        with open(self.file, "w") as file:
            file.write(json.dumps(output))



from pathlib import Path
from typing import Literal
import yaml
import shutil

from chordgen.constants import CONFIG_DIR
from chordgen.keyboards.directional import DirectionalKeyboardOptions
from chordgen.keyboards.standard import StandardKeyboardOptions
from chordgen.keyboards import Keyboard
from chordgen.output.charachorder import CharaChorderOutput
from chordgen.output.kanata import KanataOutput
from chordgen.output.qmk import QmkOutput
from chordgen.output.training import TrainingOutput
from chordgen.output.zmk import ZmkOutput
from pydantic import BaseModel, field_validator, Field
from chordgen.pydantic import File

DEFAULT_CONFIG = CONFIG_DIR / "config.yaml"
DEFAULT_CHORDS_FILE = CONFIG_DIR / "chords.csv"


class OutputOptions(BaseModel):
    formats: list[
        Literal[
            "qmk",
            "zmk",
            "charachorder",
            "kanata",
            "training",
        ]
    ] = [
        "qmk",
        "zmk",
        "charachorder",
        "kanata",
        "training",
    ]
    qmk: QmkOutput = QmkOutput()
    zmk: ZmkOutput = ZmkOutput()
    charachorder: CharaChorderOutput = CharaChorderOutput()
    kanata: KanataOutput = KanataOutput()
    training: TrainingOutput = TrainingOutput()


class KeyboardOptions(BaseModel):
    type: Literal["standard", "directional"] = "standard"
    standard: StandardKeyboardOptions = StandardKeyboardOptions()
    directional: DirectionalKeyboardOptions = DirectionalKeyboardOptions()

    _keyboard: Keyboard | None = None

    def get_keyboard(self):
        if not self._keyboard:
            self._keyboard = getattr(self, self.type).create()
        return self._keyboard


class GenOptions(BaseModel):
    file: File = DEFAULT_CHORDS_FILE
    keyboard: KeyboardOptions = KeyboardOptions()
    overwrite_alts: bool = False
    min_word_length: int = 3
    min_chord_length: int = Field(
        default=0,
        description="The minimum lenth a chord, setting this to 2 and disabling the chord key is a way to avoid needing a chord key. This works well on CharaChorder, but you will need to lower the chord timeout to avoid missfires on other keyboards.",
    )

    @field_validator("file", mode="after")
    @classmethod
    def ensure_parent_dir(cls, v: Path):
        v.parent.mkdir(parents=True, exist_ok=True)
        return v

    @field_validator("file", mode="after")
    @classmethod
    def create_file(cls, v: Path):
        if not v.exists():
            here = Path(__file__).resolve().parent
            source = here / "assets" / "chords.csv"
            shutil.copyfile(source, v)
        return v


class Config(BaseModel):
    gen: GenOptions = GenOptions()
    output: OutputOptions = OutputOptions()


def load_or_create_config(config_file: Path = DEFAULT_CONFIG) -> Config:
    if config_file != DEFAULT_CONFIG and not config_file.exists():
        raise Exception(f"Config file does not exist: {config_file}")

    config_file.parent.mkdir(parents=True, exist_ok=True)

    if config_file.exists():
        raw = yaml.safe_load(config_file.read_text()) or {}
    else:
        print(f"Creating config {config_file}")
        raw = {}
        # write defaults immediately
        default_config = Config()
        config_file.write_text(yaml.safe_dump(default_config.model_dump()))
        return default_config

    # Validate + apply defaults
    config = Config.model_validate(raw)

    # Optionally rewrite file if missing fields were added
    config_file.write_text(yaml.safe_dump(config.model_dump()))

    return config



import logging
from chordgen.chord import Chord
from chordgen.config import Config

from pattern import en


class AltGenerator:
    def __init__(self, config: Config) -> None:
        self.config = config

    def add_alt(self, chord: Chord) -> Chord:
        word = chord["word"]
        type = chord["type"]
        if self.config.overwrite_alts:
            chord["alt1"] = ""
            chord["alt2"] = ""
            chord["alt3"] = ""

        if type == "VERB":
            if not chord["alt1"]:
                chord["alt1"] = en.conjugate(word, "3sg")
            if not chord["alt2"]:
                chord["alt2"] = en.conjugate(word, "1sgp")
            if not chord["alt3"]:
                chord["alt3"] = en.conjugate(word, "part")
        elif type == "NOUN":
            if not chord["alt1"]:
                chord["alt1"] = en.pluralize(word, pos=en.NOUN)
        logging.debug(
            f"Alts for word {word}: {chord['alt1']}, {chord['alt2']}, {chord['alt3']}"
        )
        return chord



from pathlib import Path
from chordgen.chord import Chord
from chordgen.constants import CONFIG_DIR
from pydantic import BaseModel, Field
from chordgen.pydantic import File


key_map = {
    " ": "spc",
    "←": "bspc",
}


class KanataOutput(BaseModel):
    file: File = CONFIG_DIR / "kanata_chords.kbd"
    chord_keys: list[str] = ["prtsc"]
    shifted_chord_keys: list[str] = ["prtsc", "ralt"]
    alt1_keys: list[str] = ["lalt"]
    alt2_keys: list[str] = ["spc"]
    alt3_keys: list[str] = ["lalt", "spc"]
    limit: int = 0
    chord_timeout: int = 100

    key_mapping: dict[str, str] = Field(
        default={},
        description="Since Kanata combos are based in the layout in defsrc (probably qwerty) you will need to remap the letters if you are using a custom layout eg. 'a': 'b'",
    )

    def output(self, chords: list[Chord]):
        output = "(defchordsv2\n"

        def translate_macro(word):
            result = []
            for k in word:
                if k in key_map:
                    result.append(key_map[k])
                elif k.isupper():
                    result.append("S-" + k.lower())
                else:
                    result.append(k)
            return result

        def translate_chord(chord):
            result = []
            for i, k in enumerate(chord):
                if k in self.key_mapping:
                    result.append(self.key_mapping[k])
                else:
                    raise Exception(f"No key_map for {k}")
            return result

        alt_keys = [self.alt1_keys, self.alt2_keys, self.alt3_keys]
        count = 0
        for chord in chords:
            c = chord["chord"]
            if not c:
                continue

            count += 1
            if self.limit != 0 and count > self.limit:
                print(f"Stopping at line {self.limit} due to limit setting")
                break

            words = [chord["word"], chord["alt1"], chord["alt2"], chord["alt3"]]
            for i, word in enumerate(words):
                if not word:
                    continue
                alt = []
                if i > 0:
                    alt = alt_keys[i - 1]

                chord = translate_chord(c)
                macro = translate_macro(word + " ")

                output += f"  ({' '.join(self.chord_keys + alt)} {' '.join(chord)}) (macro {' '.join(macro)}) {self.chord_timeout} first-release ()\n"
                if self.shifted_chord_keys:
                    shifted_macro = translate_macro(word.capitalize() + " ")
                    output += f"  ({' '.join(self.shifted_chord_keys + alt)} {' '.join(chord)}) (macro {' '.join(shifted_macro)}) {self.chord_timeout} first-release ()\n"

        output += ")"

        print(f"Writing {self.file}")
        with open(self.file, "w") as file:
            file.write(output)



import logging
from typing import Literal
from pydantic import BaseModel, Field
import copy

FINGER_MAPPING = [
    [0, 1, 0, 0, 2, 0, 0, 3, 0, 0, 4, 0],
    [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4],
    [0, 1, 0, 0, 2, 0, 0, 3, 0, 0, 4, 0],
    [0, 5, 0],
    [5, 5, 5],
    [0, 5, 0],
    [0, 5, 0],
    [5, 5, 5],
    [0, 5, 0],
]


DIRECTION_MAPPING = [
    ["_", "u", "_", "_", "u", "_", "_", "u", "_", "_", "u", "_"],
    ["o", "c", "i", "o", "c", "i", "o", "c", "i", "o", "c", "i"],
    ["_", "d", "_", "_", "d", "_", "_", "d", "_", "_", "d", "_"],
    ["_", "u", "_"],
    ["o", "c", "i"],
    ["_", "d", "_"],
    ["_", "u", "_"],
    ["o", "c", "i"],
    ["_", "d", "_"],
]

LAYOUTS = {
    "charachorder": [
        "________________________",
        "___,_'._i__ra__l_jy_;___",
        "____u__o__e__t__n__s____",
        "_v__p_",
        "m_kf_h",
        "_c__d_",
        "____x_",
        "g_wb__",
        "_z__q_",
    ],
    "stained": [
        "_q__f__w__o__l__u__b__z_",
        "\\_'x_ky_hm_ci_gr_vp_j;_/",
        "____s__t__a__n__e__d__._",
        "______",
        "______",
        "______",
        "______",
        "______",
        "______",
    ],
    "svalboard_qwerty": [
        "_q__w__e__r__u__i__o__p_",
        "_a[_sb`dt\"fghj'yk:nl_];\\",
        "_z__x__c__v__m__,__.__/_",
        "______",
        "______",
        "______",
        "______",
        "______",
        "______",
    ],
}


class DirectionalKeyboardOptions(BaseModel):
    layout: Literal[tuple(LAYOUTS.keys()) + ("custom",)] = "charachorder"
    directional_change_penalty: int = Field(
        default=2,
        description="A penalty to add when chords have different directions per finger on the same hand. Can be set to -1 to disable this type of chord.",
    )

    custom_layout: list[str] = [
        "_X__X__X__X__X__X__X__X_",
        "X_XX_XX_XX_XX_XX_XX_XX_X",
        "_X__X__X__X__X__X__X__X_",
        "_X__X_",
        "X_XX_X",
        "_X__X_",
        "_X__X_",
        "X_XX_X",
        "_X__X_",
    ]

    effort_map: list[str] = [
        "040030020020",
        "695594493493",
        "030020010010",
        "030",
        "192",
        "040",
        "030",
        "192",
        "040",
    ]

    def create(self):
        return DirectionalKeyboard(self)


class DirectionalKeyboard:
    def __init__(self, options: DirectionalKeyboardOptions) -> None:
        if options.layout == "custom":
            self._layout = options.custom_layout
        else:
            self._layout = LAYOUTS[options.layout]

        # Break up into array of chars
        self._layout = [list(row) for row in self._layout]
        self._options = options

        # mirror maps
        effort = []
        for row in self._options.effort_map:
            r = []
            for c in row:
                if c.isnumeric():
                    r.append(int(c))
                else:
                    r.append(-1)
            # mirror
            r += r[::-1]
            effort.append(r)

        copy.deepcopy(self._options.effort_map)
        for row in effort:
            row += row[::-1]

        direction = copy.deepcopy(DIRECTION_MAPPING)
        for row in direction:
            row += row[::-1]

        finger = copy.deepcopy(FINGER_MAPPING)
        for row in finger:
            reversed = [x + 5 if x > 0 else 0 for x in row[::-1]]
            row += reversed

        hand = copy.deepcopy(FINGER_MAPPING)
        for row in hand:
            row[:] = [0] * len(row)
            right = [1] * len(row)
            row += right

        self._effort_map = {}
        self._direction_map = {}
        self._finger_map = {}
        self._hand_map = {}

        for r in range(0, len(self._layout)):
            for c in range(0, len(self._layout[r])):
                self._effort_map[self._layout[r][c]] = effort[r][c]
                self._direction_map[self._layout[r][c]] = direction[r][c]
                self._finger_map[self._layout[r][c]] = finger[r][c]
                self._hand_map[self._layout[r][c]] = hand[r][c]

    def get_directional_changes(self, chord):
        result = 0
        directions = {0: set(), 1: set()}
        for c in chord:
            directions[self._hand_map[c]].add(self._direction_map[c])

        changes = 0
        for _, hand in directions.items():
            if len(hand) > 1:
                changes += len(hand) - 1

        return result

    def score(self, chord: str) -> int:
        for i in range(0, len(chord)):
            if chord[i] not in self._effort_map:
                logging.debug(f"rejected: letter '{chord[i]}' not in keyboard layout")
                return -1

        seen = set()
        for char in chord:
            if char in seen:
                logging.debug("rejected: duplicate letters")
                return -1
            seen.add(char)

        seen = set()
        for char in chord:
            finger = self._finger_map[char]
            if finger in seen:
                logging.debug("rejected: same finger used more than once")
                return -1
            seen.add(finger)

        result = 0

        if self._options.directional_change_penalty != -1:
            changes = self.get_directional_changes(chord)
            result += changes * self._options.directional_change_penalty

        for i in range(0, len(chord)):
            result += self._effort_map[chord[i]]

        return result



import logging
import re
from chordgen.chord import load_file, validate_chords
import typer
from pathlib import Path
import jsonschema2md

from chordgen.config import DEFAULT_CONFIG, Config, load_or_create_config
from chordgen.gen import gen as run_gen


app = typer.Typer()


class State:
    config: Config


@app.callback()
def callback(
    ctx: typer.Context,
    config: Path = typer.Option(
        DEFAULT_CONFIG, "-c", "--config", help="Path to config file"
    ),
):
    logging.basicConfig(level="INFO")
    if ctx.invoked_subcommand != "setup":
        if not config.exists():
            print(
                f"Error: config {config} does not exist, run 'chordgen setup' to create it"
            )
            raise typer.Abort()

    loaded_config = load_or_create_config(config)
    if ctx.invoked_subcommand != "setup":
        if not loaded_config.gen.file.exists():
            print(
                f"Error: chord file {loaded_config.gen.file} does not exist, run 'chordgen setup' to create it"
            )
            raise typer.Abort()

    State.config = loaded_config


@app.command()
def setup():
    print("Setup Complete")


@app.command()
def gen():
    run_gen(State.config.gen)


@app.command()
def output():
    chords = load_file(State.config.gen.file)
    validate_chords(chords)

    for format in State.config.output.formats:
        print(f"Running output format '{format}'")
        format = getattr(State.config.output, format)
        format.output(chords)


@app.command()
def schema():
    parser = jsonschema2md.Parser()
    md = "".join(parser.parse_schema(Config.model_json_schema()))
    pattern = r"/(?:Users|home)/[^/]+/"
    md = re.sub(pattern, "~/", md)

    with open("schema.md", "w") as f:
        f.write(md)


if __name__ == "__main__":
    app()



import logging
from typing import Literal
from pydantic import BaseModel, Field
import copy

FINGER_MAPPING = [
    [1, 1, 2, 3, 4, 4, 5, 5, 6, 7, 8, 8],
    [1, 1, 2, 3, 4, 4, 5, 5, 6, 7, 8, 8],
    [1, 1, 2, 3, 4, 4, 5, 5, 6, 7, 8, 8],
    [9, 9, 0, 0],
]


HAND_ROW_MAPPING = [
    ["tl", "tl", "tl", "tl", "tl", "tr", "tr", "tr", "tr", "tr", "tr", "tr"],
    ["ml", "ml", "ml", "ml", "ml", "mr", "mr", "mr", "mr", "mr", "mr", "mr"],
    ["bl", "bl", "bl", "bl", "bl", "br", "br", "br", "br", "br", "br", "br"],
    ["th", "th", "th", "th"],
]

BANNED_CHORDS = [
    # Index same row only
    [
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0],
        [0, 0],
    ],
    [
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 1],
        [0, 0],
    ],
    [
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1],
        [0, 0],
    ],
    [
        [0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0],
    ],
    [
        [0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0],
        [0, 0],
    ],
    [
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 1],
        [0, 0, 0, 0, 1, 0],
        [0, 0],
    ],
    # Pinky same row only
    [
        [0, 1, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0],
    ],
    [
        [0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0],
        [0, 0],
    ],
    [
        [0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0],
        [0, 0],
    ],
    [
        [1, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 0],
    ],
    [
        [1, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0],
    ],
    [
        [0, 0, 0, 0, 0, 0],
        [1, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0],
    ],
    # Uncomfortable
    [
        [0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 1],
        [0, 0, 0, 1, 0, 0],
        [0, 0],
    ],
    [
        [0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0],
        [0, 0],
    ],
]

LAYOUTS = {
    "qwerty": [
        "_qwertyuiop_",
        "_asdfghjkl;_",
        "_zxcvbnm,./_",
        "____",
    ],
    "colemak": [
        "_qwfpgjluy;_",
        "_arstdhneio_",
        "_zxcvbkm,./_",
        "____",
    ],
    "colemak_dh": [
        "_qwfpbjluy;_",
        "_arstgmneio_",
        "_zxcdvkh,./_",
        "____",
    ],
    "canary": [
        "_wlypbzfou'_",
        "_crstgmneia_",
        "_qjvdkxh;,._",
        "____",
    ],
    "engram_2021": [
        "_byou'\"ldvwz",
        "_ciea,.htsnq",
        "_gxjk-?rmfp_",
        "____",
    ],
    "engram_en": [
        "_byou'\"dngvq",
        "_hiae,.trscz",
        "_kjxw-?mlfp_",
        "____",
    ],
    "enthium_v14": [
        "_qyou=xldpz_",
        "bciae-khtsnw",
        "_,,.;/jmgfv_",
        "__r_",
    ],
}


class StandardKeyboardOptions(BaseModel):
    layout: Literal[tuple(LAYOUTS.keys()) + ("custom",)] = "qwerty"
    scissor_penalty: int = Field(
        default=3,
        description="A penalty to add when pressing keys on the top and bottom rows together. Can be set to -1 to disable this type of chord.",
    )
    same_column_chord_penalty: int = Field(
        default=2,
        description="A penalty to add when pressing 2 keys and the same time with the same finger in the same column. Can be set to -1 to disable this type of chord.",
    )
    same_row_chord_penalty: int = Field(
        default=2,
        description="A penalty to add when pressing 2 keys and the same time with the same finger in the same row. Can be set to -1 to disable this type of chord.",
    )

    custom_layout: list[str] = [
        "_qwertyuiop_",
        "_asdfghjkl;_",
        "_zxcvbnm,./_",
        "____",
    ]

    effort_map: list[str] = [
        "965446",
        "732116",
        "865536",
        "43",
    ]

    def create(self):
        return StandardKeyboard(self)


class StandardKeyboard:
    def __init__(self, options: StandardKeyboardOptions) -> None:
        if options.layout == "custom":
            self._layout = options.custom_layout
        else:
            self._layout = LAYOUTS[options.layout]

        # Break up into array of chars
        self._layout = [list(row) for row in self._layout]
        self._options = options

        effort = []
        for row in self._options.effort_map:
            r = []
            for c in row:
                if c.isnumeric():
                    r.append(int(c))
                else:
                    r.append(-1)
            # mirror
            r += r[::-1]
            effort.append(r)

        # self._finger_map = {}
        self._effort_map = {}
        self._hand_row_map = {}
        self._banned_chords_sets = []
        self._chord_map = copy.deepcopy(self._layout)
        self._chord_lookup = {}

        for r in range(0, len(self._layout)):
            for c in range(0, len(self._layout[r])):
                # self._finger_map[self._layout[r][c]] = FINGER_MAPPING[r][c]
                self._effort_map[self._layout[r][c]] = effort[r][c]
                self._hand_row_map[self._layout[r][c]] = HAND_ROW_MAPPING[r][c]
                self._chord_map[r][c] = 0
                self._chord_lookup[self._layout[r][c]] = (r, c)

        # Add mirrored chords and padding
        mirrored = []
        # banned chords, this left hand side is mirrored
        banned_chords = copy.deepcopy(BANNED_CHORDS)
        padding = [0, 0, 0, 0, 0, 0]
        for ban in banned_chords:
            mirror = []
            for r in range(0, len(ban)):
                mirror.append(padding + list(reversed(ban[r])))
                ban[r] += padding
            mirrored.append(mirror)
        banned_chords += mirrored

        for ban in banned_chords:
            s = set()
            for r in range(0, len(ban)):
                for c in range(0, len(ban[r])):
                    if ban[r][c]:
                        s.add(self._layout[r][c])
            self._banned_chords_sets.append(s)

    def get_scissor_count(self, chord):
        result = 0
        indexes = {
            "tl": 0,
            "ml": 0,
            "bl": 0,
            "tr": 0,
            "mr": 0,
            "br": 0,
            "th": 0,
        }
        for i in range(0, len(chord)):
            indexes[self._hand_row_map[chord[i]]] += 1

        if indexes["tl"] and indexes["bl"]:
            result += min(indexes["tl"], indexes["bl"])

        if indexes["tr"] and indexes["br"]:
            result += min(indexes["tr"], indexes["br"])

        return result

    def get_chord_map(self, chord: str):
        map = copy.deepcopy(self._chord_map)
        for c in chord:
            offset = self._chord_lookup[c]
            map[offset[0]][offset[1]] = 1
        return map

    def get_same_column_chord(self, chord_map) -> int:
        count = 0
        for i in range(0, len(chord_map[0])):
            if chord_map[0][i] and chord_map[1][i]:
                if self._options.same_column_chord_penalty == -1:
                    return -1

                count += 1

            # Top and bottom row on same finger is excluded
            if chord_map[0][i] and chord_map[2][i]:
                return -1

        return count

    def get_same_row_chord(self, chord_map) -> int:
        count = 0
        for r in range(0, len(chord_map)):
            for c in range(0, len(chord_map[r]) - 1):
                if (
                    chord_map[r][c]
                    and chord_map[r][c + 1]
                    and FINGER_MAPPING[r][c] == FINGER_MAPPING[r][c + 1]
                ):
                    if self._options.same_row_chord_penalty == -1:
                        return -1
                    count += 1

        return count

    def score(self, chord: str) -> int:
        for i in range(0, len(chord)):
            if chord[i] not in self._effort_map:
                logging.debug(f"rejected: letter '{chord[i]}' not in keyboard layout")
                return -1

        seen = set()
        for char in chord:
            if char in seen:
                logging.debug("rejected: duplicate letters")
                return -1
            seen.add(char)

        for ban in self._banned_chords_sets:
            if ban.issubset(seen):
                logging.debug("rejected: banned chord")
                return -1

        chord_map = self.get_chord_map(chord)
        result = 0
        same_column_chord = self.get_same_column_chord(chord_map)
        if same_column_chord == -1:
            logging.debug("rejected: same column chord")
            return -1

        result += same_column_chord * self._options.same_column_chord_penalty

        same_row_chord = self.get_same_row_chord(chord_map)
        if same_row_chord == -1:
            logging.debug("rejected: same column chord")
            return -1

        result += same_row_chord * self._options.same_row_chord_penalty

        scissor_count = self.get_scissor_count(chord)
        if scissor_count:
            if self._options.scissor_penalty == -1:
                logging.debug("rejected: scissor")
                return -1
            result += self._options.scissor_penalty * scissor_count

        for i in range(0, len(chord)):
            result += self._effort_map[chord[i]]

        return result



import csv
import logging
from concurrent.futures import ProcessPoolExecutor
from tqdm import tqdm

from chordgen.alt_generator import AltGenerator
from chordgen.config import Config, GenOptions
from chordgen.scorer import Scorer


def gen(options: GenOptions) -> None:
    scorer = Scorer(options)
    with open(options.file) as f:
        reader = csv.DictReader(f)
        print("Finding and scoring chords")
        chords = [line for line in reader]
        if len(chords) == 0:
            raise Exception("No rows found in chords file")
        with ProcessPoolExecutor() as executor:
            chords = list(
                tqdm(
                    executor.map(scorer.score, chords, chunksize=10),
                    total=len(chords),
                )
            )

    used = {}
    seen = {}
    no_options = []
    duplicate = []

    print("Setting reserved chords")
    for chord in tqdm(chords):
        reserved_chord = chord["reserved_chord"]
        chord["chord"] = reserved_chord
        if not reserved_chord:
            continue
        sorted_chord = "".join(sorted(reserved_chord))
        if sorted_chord in used:
            raise Exception(
                f"Reserved chord for word {chord['word']} already used for {used[sorted_chord]['word']}"
            )
        used[sorted_chord] = chord

    print("Selecting chords")
    for chord in tqdm(chords):
        word = chord["word"].lower()
        if word in seen:
            duplicate.append(word)
            continue
        seen[word] = True
        if len(word) < options.min_word_length:
            continue
        reserved_chord = chord["reserved_chord"]
        if reserved_chord:
            continue
        for option in chord.get("options", []):
            if len(option["chord"]) < options.min_chord_length:
                continue
            sorted_chord = "".join(sorted(option["chord"]))
            if sorted_chord not in used:
                chord["chord"] = option["chord"]
                used[sorted_chord] = chord
                break

        if not chord["chord"]:
            no_options.append(word)

    if len(no_options) > 0:
        print(
            f"Unable to find any options for {len(no_options)} words: {', '.join(no_options)}"
        )
    if len(duplicate) > 0:
        print(f"Ignored {len(duplicate)} duplicate words: {', '.join(duplicate)}")

    print("Generating alts")
    alt_generator = AltGenerator(options)
    with ProcessPoolExecutor() as executor:
        chords = list(
            tqdm(
                executor.map(alt_generator.add_alt, chords, chunksize=10),
                total=len(chords),
            )
        )

    print(f"Writing {options.file}")
    with open(options.file, "w", newline="") as f:
        fieldnames = list(chords[0].keys())
        if "options" in fieldnames:
            fieldnames.remove("options")
        writer = csv.DictWriter(
            f, fieldnames=fieldnames, extrasaction="ignore", lineterminator="\n"
        )
        writer.writeheader()
        writer.writerows(chords)



import logging
from chordgen.chord import Chord, Option
from chordgen.config import GenOptions
from chordgen.utils import find_combinations


class Scorer:
    def __init__(self, options: GenOptions) -> None:
        self._options = options
        self._keyboard = options.keyboard.get_keyboard()

    def score(self, chord: Chord) -> Chord:
        if chord["reserved_chord"]:
            return chord

        if len(chord["word"]) < self._options.min_word_length:
            return chord

        combinations = find_combinations(chord["word"].lower())
        scores = [self._keyboard.score(combination) for combination in combinations]
        options: list[Option] = [
            {"chord": combination, "score": scores[i]}
            for i, combination in enumerate(combinations)
            if scores[i] != -1
        ]
        options = sorted(options, key=lambda x: x["score"])
        chord["options"] = options
        logging.debug(f"Computed options for {chord['word']}")
        return chord
